from radical.entk import Pipeline, Stage, Task, AppManager, ResourceManager
from radical.entk import Profiler
import os

# ------------------------------------------------------------------------------
# Set default verbosity
if os.environ.get('RADICAL_ENTK_VERBOSE') == None:
    os.environ['RADICAL_ENTK_REPORT'] = 'True'


os.environ['RADICAL_ENTK_PROFILE'] = 'True'

if __name__ == '__main__':

    # Create a Pipeline object
    p = Pipeline()

    # Create a Stage object 
    s1 = Stage()

    # List to hold uids of Tasks of Stage 1
    s1_task_uids = list()

    for cnt in range(10):

        # Create a Task object
        t = Task()
        t.executable = ['/bin/echo']   # Assign executable to the task   
        t.arguments = ['I am task %s in %s'%(cnt, s1.name)]  # Assign arguments for the task executable

        # Add the Task to the Stage
        s1.add_tasks(t)

        # Add Task uid to list
        s1_task_uids.append(t.uid)

    # Add Stage to the Pipeline
    p.add_stages(s1)


    # Create another Stage object
    s2 = Stage()

    # List to hold uids of Tasks of Stage 2
    s2_task_uids = list()

    for cnt in range(5):

        # Create a Task object
        t = Task()
        t.executable = ['/bin/echo']   # Assign executable to the task   
        t.arguments = ['I am task %s in %s'%(cnt, s2.name)]  # Assign arguments for the task executable

        # Add the Task to the Stage
        s2.add_tasks(t)

        # Add Task uid to list
        s2_task_uids.append(t.uid)

    # Add Stage to the Pipeline
    p.add_stages(s2)


    # Create a dictionary describe four mandatory keys:
    # resource, walltime, cpus and project
    # resource is 'local.localhost' to execute locally
    res_dict = {

            'resource': 'local.localhost',
            'walltime': 10,
            'cpus': 1,
            'project': '',
    }

    # Create Resource Manager object with the above resource description
    rman = ResourceManager(res_dict)

    # Create Application Manager
    appman = AppManager()

    # Assign resource manager to the Application Manager
    appman.resource_manager = rman

    # Assign the workflow as a set of Pipelines to the Application Manager
    appman.assign_workflow(set([p]))

    # Run the Application Manager
    appman.run()


    p = Profiler(src='./%s'%appman.sid)

    print 'Tasks in Stage 1: ', s1_task_uids
    print 'Execution time: ', p.duration(objects = s1_task_uids, states=['SCHEDULING', 'EXECUTED'])

    print 'Tasks in Stage 2: ', s2_task_uids
    print 'Execution time: ', p.duration(objects = s2_task_uids, states=['SCHEDULING', 'EXECUTED'])



import radical.pilot as rp
from radical.entk import Task
import radical.utils as ru
import traceback
from radical.entk.exceptions import *
import os

logger = ru.Logger('radical.entk.task_processor')


def resolve_placeholders(path, placeholder_dict):
    """
    **Purpose**: Substitute placeholders in staging attributes of a Task with actual paths to the corresponding tasks.

    :arguments:
        :path: string describing the staging paths, possibly containing a placeholder
        :placeholder_dict: dictionary holding the values for placeholders

    """

    try:

        if isinstance(path, unicode):
            path = str(path)

        if not isinstance(path, str):
            raise TypeError(expected_type=str, actual_type=type(path))

        if '$' not in path:
            return path

        # Extract placeholder from path
        if len(path.split('>')) == 1:
            placeholder = path.split('/')[0]
        else:
            if path.split('>')[0].strip().startswith('$'):
                placeholder = path.split('>')[0].strip().split('/')[0]
            else:
                placeholder = path.split('>')[1].strip().split('/')[0]

        # SHARED
        if placeholder == "$SHARED":
            return path.replace(placeholder, 'pilot://')

        # Expected placeholder format:
        # $Pipeline_{pipeline.uid}_Stage_{stage.uid}_Task_{task.uid}

        broken_placeholder = placeholder.split('/')[0].split('_')

        if not len(broken_placeholder) == 6:
            raise ValueError(
                obj='placeholder',
                attribute='length',
                expected_value='$Pipeline_{pipeline.uid}_Stage_{stage.uid}_Task_{task.uid} or $SHARED',
                actual_value=broken_placeholder)

        pipeline_name = broken_placeholder[1]
        stage_name = broken_placeholder[3]
        task_name = broken_placeholder[5]

        if pipeline_name in placeholder_dict.keys():
            if stage_name in placeholder_dict[pipeline_name].keys():
                if task_name in placeholder_dict[pipeline_name][stage_name].keys():
                    resolved_placeholder = path.replace(placeholder, placeholder_dict[
                                                        pipeline_name][stage_name][task_name]['path'])
                else:
                    logger.warning('%s not assigned to any task in Stage %s Pipeline %s' %
                                   (task_name, stage_name, pipeline_name))
            else:
                logger.warning('%s not assigned to any Stage in Pipeline %s' % (
                    stage_name, pipeline_name))
        else:
            logger.warning('%s not assigned to any Pipeline' % (pipeline_name))

        return resolved_placeholder

    except Exception, ex:

        logger.error('Failed to resolve placeholder %s, error: %s' %
                     (path, ex))
        raise


def resolve_arguments(args, placeholder_dict):

    resolved_args = list()

    for entry in args:

        # If entry starts with $, it has a placeholder
        # and needs to be resolved based after a lookup in
        # the placeholder_dict
        if isinstance(entry, str) and entry.startswith('$'):

            placeholder = entry.split('/')[0]

            if placeholder == "$SHARED":
                entry = entry.replace(placeholder, '$RP_PILOT_STAGING')

            else:
                broken_placeholder = placeholder.split('_')

                if not len(broken_placeholder) == 6:
                    raise ValueError(
                        obj='placeholder',
                        attribute='length',
                        expected_value='$Pipeline_{pipeline.uid}_Stage_{stage.uid}_Task_{task.uid} or $SHARED',
                        actual_value=broken_placeholder)

                pipeline_name = broken_placeholder[1]
                stage_name = broken_placeholder[3]
                task_name = broken_placeholder[5]

                try:
                    entry = entry.replace(
                        placeholder, placeholder_dict[pipeline_name][stage_name][task_name]['path'])

                except Exception as ex:
                    logger.warning('Argument parsing failed. Task %s of Stage %s in Pipeline %s does not exist' %
                                   (task_name, stage_name, pipeline_name))

        resolved_args.append(entry)

    return resolved_args


def get_input_list_from_task(task, placeholder_dict):
    """
    Purpose: Parse a Task object to extract the files to be staged as the output. 

    Details: The extracted data is then converted into the appropriate RP directive depending on whether the data
    is to be copied/downloaded.

    :arguments: 
        :task: EnTK Task object
        :placeholder_dict: dictionary holding the values for placeholders

    :return: list of RP directives for the files that need to be staged out
    """

    try:

        if not isinstance(task, Task):
            raise TypeError(expected_type=Task, actual_type=type(task))

        input_data = []

        if task.link_input_data:

            for path in task.link_input_data:

                path = resolve_placeholders(path, placeholder_dict)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.LINK
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.LINK
                    }
                input_data.append(temp)

        if task.upload_input_data:

            for path in task.upload_input_data:

                path = resolve_placeholders(path, placeholder_dict)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip()
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip())
                    }
                input_data.append(temp)

        if task.copy_input_data:

            for path in task.copy_input_data:

                path = resolve_placeholders(path, placeholder_dict)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.COPY
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.COPY
                    }
                input_data.append(temp)

        if task.move_input_data:

            for path in task.move_input_data:

                path = resolve_placeholders(path, placeholder_dict)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.MOVE
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.MOVE
                    }

                input_data.append(temp)

        return input_data

    except Exception, ex:

        logger.error(
            'Failed to get input list of files from task, error: %s' % ex)
        raise


def get_output_list_from_task(task, placeholder_dict):
    """
    Purpose: Parse a Task object to extract the files to be staged as the output. 

    Details: The extracted data is then converted into the appropriate RP directive depending on whether the data
    is to be copied/downloaded.

    :arguments: 
        :task: EnTK Task object
        :placeholder_dict: dictionary holding the values for placeholders

    :return: list of RP directives for the files that need to be staged out

    """

    try:

        if not isinstance(task, Task):
            raise TypeError(expected_type=Task, actual_type=type(task))

        output_data = []

        if task.copy_output_data:

            for path in task.copy_output_data:

                path = resolve_placeholders(path, placeholder_dict)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.COPY
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.COPY
                    }
                output_data.append(temp)

        if task.download_output_data:

            for path in task.download_output_data:

                path = resolve_placeholders(path, placeholder_dict)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip()
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip())
                    }
                output_data.append(temp)

        if task.move_output_data:

            for path in task.move_output_data:

                path = resolve_placeholders(path, placeholder_dict)

                if len(path.split('>')) > 1:

                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': path.split('>')[1].strip(),
                        'action': rp.MOVE
                    }
                else:
                    temp = {
                        'source': path.split('>')[0].strip(),
                        'target': os.path.basename(path.split('>')[0].strip()),
                        'action': rp.MOVE
                    }
                    
                output_data.append(temp)


        return output_data

    except Exception, ex:
        logger.error(
            'Failed to get output list of files from task, error: %s' % ex)
        raise


def create_cud_from_task(task, placeholder_dict, prof=None):
    """
    Purpose: Create a Compute Unit description based on the defined Task.

    :arguments: 
        :task: EnTK Task object
        :placeholder_dict: dictionary holding the values for placeholders

    :return: ComputeUnitDescription
    """

    try:

        logger.debug('Creating CU from Task %s' % (task.uid))

        if prof:
            prof.prof('cud from task - create', uid=task.uid)

        cud = rp.ComputeUnitDescription()
        cud.name = '%s,%s,%s,%s,%s,%s' % (task.uid, task.name,
                                          task.parent_stage['uid'], task.parent_stage['name'],
                                          task.parent_pipeline['uid'], task.parent_pipeline['name'])
        cud.pre_exec = task.pre_exec
        cud.executable = task.executable
        cud.arguments = resolve_arguments(task.arguments, placeholder_dict)
        cud.post_exec = task.post_exec
        if task.tag:
            cud.tag = resolve_tags(task.tag, placeholder_dict)

        cud.cpu_processes = task.cpu_reqs['processes']
        cud.cpu_threads = task.cpu_reqs['threads_per_process']
        cud.cpu_process_type = task.cpu_reqs['process_type']
        cud.cpu_thread_type = task.cpu_reqs['thread_type']
        cud.gpu_processes = task.gpu_reqs['processes']
        cud.gpu_threads = task.gpu_reqs['threads_per_process']
        cud.gpu_process_type = task.gpu_reqs['process_type']
        cud.gpu_thread_type = task.gpu_reqs['thread_type']
        if task.lfs_per_process:
            cud.lfs_per_process = task.lfs_per_process

        cud.input_staging = get_input_list_from_task(task, placeholder_dict)
        cud.output_staging = get_output_list_from_task(task, placeholder_dict)

        if prof:
            prof.prof('cud from task - done', uid=task.uid)

        logger.debug('CU %s created from Task %s' % (cud.name, task.uid))

        return cud

    except Exception, ex:
        logger.error('CU creation failed, error: %s' % ex)
        raise


def create_task_from_cu(cu, prof=None):
    """
    Purpose: Create a Task based on the Compute Unit.

    Details: Currently, only the uid, parent_stage and parent_pipeline are retrieved. The exact initial Task (that was
    converted to a CUD) cannot be recovered as the RP API does not provide the same attributes for a CU as for a CUD. 
    Also, this is not required for the most part.

    TODO: Add exit code, stdout, stderr and path attributes to a Task. These can be extracted from a CU

    :arguments: 
        :cu: RP Compute Unit

    :return: Task
    """

    try:

        logger.debug('Create Task from CU %s' % cu.name)

        if prof:
            prof.prof('task from cu - create',
                      uid=cu.name.split(',')[0].strip())

        task = Task()
        task.uid = cu.name.split(',')[0].strip()
        task.name = cu.name.split(',')[1].strip()
        task.parent_stage['uid'] = cu.name.split(',')[2].strip()
        task.parent_stage['name'] = cu.name.split(',')[3].strip()
        task.parent_pipeline['uid'] = cu.name.split(',')[4].strip()
        task.parent_pipeline['name'] = cu.name.split(',')[5].strip()
        task.rts_uid = cu.uid

        if cu.exit_code is not None:
            task.exit_code = cu.exit_code
        else:

            if cu.state == rp.DONE:
                task.exit_code = 0
            else:
                task.exit_code = 1

        task.path = ru.Url(cu.sandbox).path

        if prof:
            prof.prof('task from cu - done', uid=cu.name.split(',')[0].strip())

        logger.debug('Task %s created from CU %s' % (task.uid, cu.name))

        return task

    except Exception, ex:
        logger.error('Task creation from CU failed, error: %s' % ex)
        raise



import radical.utils as ru
from radical.entk.exceptions import *
from radical.entk import states


class Task(object):

    """
    A Task is an abstraction of a computational unit. In this case, a Task
    consists of its executable along with its required software environment,
    files to be staged as input and output.

    At the user level a Task is to be populated by assigning attributes.
    Internally, an empty Task is created and population using the `from_dict`
    function. This is to avoid creating Tasks with new `uid` as tasks with new
    `uid` offset the uid count file in radical.utils and can potentially affect
    the profiling if not taken care.
    """

    def __init__(self):

        self._uid = None
        self._name = None

        self._state = states.INITIAL

        # Attributes necessary for execution
        self._pre_exec = list()
        self._executable = list()
        self._arguments = list()
        self._post_exec = list()
        self._cpu_reqs = {'processes': 1,
                          'process_type': None,
                          'threads_per_process': 1,
                          'thread_type': None
                          }
        self._gpu_reqs = {'processes': 0,
                          'process_type': None,
                          'threads_per_process': 0,
                          'thread_type': None
                          }
        self._lfs_per_process = 0

        # Data staging attributes
        self._upload_input_data = list()
        self._copy_input_data = list()
        self._link_input_data = list()
        self._move_input_data = list()
        self._copy_output_data = list()
        self._move_output_data = list()
        self._download_output_data = list()

        self._path = None
        self._exit_code = None
        self._tag = None

        # Keep track of states attained
        self._state_history = [states.INITIAL]

        # The following help in updation
        # Stage this task belongs to
        self._p_stage = {'uid': None, 'name': None}
        # Pipeline this task belongs to
        self._p_pipeline = {'uid': None, 'name': None}

    # ------------------------------------------------------------------------------------------------------------------
    # Getter functions
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def uid(self):
        """
        Unique ID of the current task

        :getter: Returns the unique id of the current task
        :type: String
        """
        return self._uid

    @property
    def name(self):
        """
        Name of the task. Do not use a ',' or '_' in an object's name.

        :getter: Returns the name of the current task
        :setter: Assigns the name of the current task
        :type: String
        """

        return self._name

    @property
    def state(self):
        """
        Current state of the task

        :getter: Returns the state of the current task
        :type: String
        """

        return self._state

    @property
    def pre_exec(self):
        """
        List of commands to be executed prior to the executable

        :getter: return the list of commands
        :setter: assign the list of commands
        :arguments: list of strings
        """
        return self._pre_exec

    @property
    def executable(self):
        """
        A unix-based kernel to be executed

        :getter: returns the executable of the current task
        :setter: assigns the executable for the current task
        :arguments: string
        """
        return self._executable

    @property
    def arguments(self):
        """
        List of arguments to be supplied to the executable

        :getter: returns the list of arguments of the current task
        :setter: assigns a list of arguments to the current task
        :arguments: list of strings
        """
        return self._arguments

    @property
    def post_exec(self):
        """
        List of commands to be executed post executable

        :getter: return the list of commands
        :setter: assign the list of commands
        :arguments: list of strings
        """

        return self._post_exec

    @property
    def cpu_reqs(self):
        """
        **Purpose:** The CPU requirements of the current Task.

        The requirements are described in terms of the number of processes and threads to
        be run in this Task. The expected format is:

        task.cpu_reqs = {
                            |  'processes': X,
                            |  'process_type': None/MPI,
                            |  'threads_per_process': Y,
                            |  'thread_type': None/OpenMP
                        }

        This description means that the Task is going to spawn X processes and Y threads
        per each of these processes to run on CPUs. Hence, the total number of cpus required by the
        Task is X*Y for all the processes and threads to execute concurrently. The
        same assumption is made in implementation and X*Y cpus are requested for this
        Task.

        The default value is:

        task.cpu_reqs = {
                            |  'processes': 1,
                            |  'process_type': None,
                            |  'threads_per_process': 1,
                            |  'thread_type': None
                        }

        This description requests 1 core and expected the executable to non-MPI and
        single threaded.

        :getter: return the cpu requirement of the current Task
        :setter: assign the cpu requirement of the current Task
        :arguments: dict

        """

        return self._cpu_reqs

    @property
    def gpu_reqs(self):
        """
        **Purpose:** The GPU requirements of the current Task.

        The requirements are described in terms of the number of processes and threads to
        be run in this Task. The expected format is:

        task.gpu_reqs = {
                            |  'processes': X,
                            |  'process_type': None/MPI,
                            |  'threads_per_process': Y,
                            |  'thread_type': None/OpenMP
                        }

        This description means that the Task is going to spawn X processes and Y threads
        per each of these processes to run on GPUs. Hence, the total number of gpus required by the
        Task is X*Y for all the processes and threads to execute concurrently. The
        same assumption is made in implementation and X*Y gpus are requested for this
        Task.

        The default value is:

        task.gpu_reqs = {
                            |  'processes': 0,
                            |  'process_type': None,
                            |  'threads_per_process': 0,
                            |  'thread_type': None
                        }

        This description requests 0 gpus as not all machines have GPUs.

        :getter: return the gpu requirement of the current Task
        :setter: assign the gpu requirement of the current Task
        :arguments: dict

        """

        return self._gpu_reqs

    @property
    def lfs_per_process(self):
        """
        Set the amount of local file-storage space required by the task
        """
        return self._lfs_per_process

    @property
    def upload_input_data(self):
        """
        List of files to be transferred from local machine to the location of the current task
        on the remote machine

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        """

        return self._upload_input_data

    @property
    def copy_input_data(self):
        """
        List of files to be copied from a location on the remote machine to the location of
        current task on the remote machine

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        """

        return self._copy_input_data

    @property
    def link_input_data(self):
        """
        List of files to be linked from a location on the remote machine to the location of
        current task on the remote machine

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        """

        return self._link_input_data


    @property
    def move_input_data(self):
        """
        List of files to be move from a location on the remote machine to the location of
        current task on the remote machine

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        """

        return self._move_input_data

    @property
    def copy_output_data(self):
        """
        List of files to be copied from the location of the current task to another location
        on the remote machine

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        """

        return self._copy_output_data


    @property
    def move_output_data(self):
        """
        List of files to be copied from the location of the current task to another location
        on the remote machine

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        """

        return self._move_output_data

    @property
    def download_output_data(self):
        """
        List of files to be downloaded from the location of the current task to a location
        on the local machine.

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        """
        return self._download_output_data

    @property
    def exit_code(self):
        """
        Get the exit code for DONE tasks. 0 for successful tasks, 1 for failed tasks.

        :getter: return the exit code of the current task
        """
        return self._exit_code

    @property
    def path(self):
        """
        Get the path of the task on the remote machine. Useful to reference files
        generated in the current task.

        :getter: return the path of the current task
        """

        return self._path

    @property
    def tag(self):
        """
        Set the tag for the task that can be used while scheduling by the RTS

        :getter: return the tag of the current task
        """

        return self._tag

    @property
    def parent_stage(self):
        """
        :getter: Returns the stage this task belongs to
        :setter: Assigns the stage uid this task belongs to
        """
        return self._p_stage

    @property
    def parent_pipeline(self):
        """
        :getter: Returns the pipeline this task belongs to
        :setter: Assigns the pipeline uid this task belongs to
        """

        return self._p_pipeline

    @property
    def state_history(self):
        """
        Returns a list of the states obtained in temporal order

        :return: list
        """

        return self._state_history

    # ------------------------------------------------------------------------------------------------------------------
    # Setter functions
    # ------------------------------------------------------------------------------------------------------------------

    @uid.setter
    def uid(self, val):
        if isinstance(val, str):
            self._uid = val
        else:
            raise TypeError(expected_type=str, actual_type=type(val))

    @name.setter
    def name(self, value):
        if isinstance(value, str):
            if ',' in value:
                raise Error(
                    "Using ',' or '_' in an object's name may corrupt the profiling and internal mapping tables")
            else:
                self._name = value
        else:
            raise TypeError(expected_type=str, actual_type=type(value))

    @state.setter
    def state(self, value):
        if isinstance(value, str):
            if value in states._task_state_values.keys():
                self._state = value
                self._state_history.append(value)
            else:
                raise ValueError(obj=self._uid,
                                 attribute='state',
                                 expected_value=states._task_state_values.keys(),
                                 actual_value=value)
        else:
            raise TypeError(expected_type=str, actual_type=type(val))

    @pre_exec.setter
    def pre_exec(self, val):
        if isinstance(val, list):
            self._pre_exec = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))

    @executable.setter
    def executable(self, val):
        if isinstance(val, list):
            self._executable = val
        elif isinstance(val, str):
            self._executable = [val]
        else:
            raise TypeError(expected_type='list or str', actual_type=type(val))

    @arguments.setter
    def arguments(self, val):
        if isinstance(val, list):
            self._arguments = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))

    @post_exec.setter
    def post_exec(self, val):
        if isinstance(val, list):
            self._post_exec = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))

    @cpu_reqs.setter
    def cpu_reqs(self, val):
        if isinstance(val, dict):

            expected_keys = set(
                ['processes', 'threads_per_process', 'process_type', 'thread_type'])

            if set(val.keys()) <= expected_keys:

                if type(val.get('processes')) in [type(None), int]:
                    self._cpu_reqs['processes'] = val.get('processes', 1)
                else:
                    raise TypeError(expected_type=int,
                                    actual_type=type(val.get('processes')),
                                    entity='processes'
                                    )

                if val.get('process_type') in [None, 'MPI', '']:
                    self._cpu_reqs['process_type'] = val.get('process_type')
                else:
                    raise ValueError(expected_value='None or MPI',
                                     actual_value=val.get('process_type'),
                                     obj='cpu_reqs',
                                     attribute='process_type'
                                     )

                if type(val.get('threads_per_process')) in [type(None), int]:
                    self._cpu_reqs['threads_per_process'] = val.get(
                        'threads_per_process', 1)
                else:
                    raise TypeError(expected_type=int,
                                    actual_type=type(
                                        val.get('threads_per_process')),
                                    entity='threads_per_process'
                                    )

                if val.get('thread_type') in [None, 'OpenMP', '']:
                    self._cpu_reqs['thread_type'] = val.get('thread_type')
                else:
                    raise ValueError(expected_value='None or OpenMP',
                                     actual_value=val.get('thread_type'),
                                     obj='cpu_reqs',
                                     attribute='thread_type'
                                     )

            else:
                raise MissingError(
                    obj='cpu_reqs', missing_attribute=expected_keys - set(val.keys()))

    @gpu_reqs.setter
    def gpu_reqs(self, val):
        if isinstance(val, dict):

            expected_keys = set(
                ['processes', 'threads_per_process', 'process_type', 'thread_type'])

            if set(val.keys()) <= expected_keys:

                if type(val.get('processes')) in [type(None), int]:
                    self._gpu_reqs['processes'] = val.get('processes', 1)
                else:
                    raise TypeError(expected_type=dict,
                                    actual_type=type(val.get('processes')),
                                    entity='processes'
                                    )

                if val.get('process_type') in [None, 'MPI', '']:
                    self._gpu_reqs['process_type'] = val.get('process_type')
                else:
                    raise ValueError(expected_value='None or MPI',
                                     actual_value=val.get('process_type'),
                                     obj='gpu_reqs',
                                     attribute='process_type'
                                     )

                if type(val.get('threads_per_process')) in [type(None), int]:
                    self._gpu_reqs['threads_per_process'] = val.get(
                        'threads_per_process', 1)
                else:
                    raise TypeError(expected_type=int,
                                    actual_type=type(
                                        val.get('threads_per_process')),
                                    entity='threads_per_process'
                                    )

                if val.get('thread_type') in [None, 'OpenMP', '']:
                    self._gpu_reqs['thread_type'] = val.get('thread_type')
                else:
                    raise ValueError(expected_value='None or OpenMP',
                                     actual_value=val.get('thread_type'),
                                     obj='gpu_reqs',
                                     attribute='thread_type'
                                     )

            else:
                raise MissingError(
                    obj='gpu_reqs', missing_attribute=expected_keys - set(val.keys()))

    @lfs_per_process.setter
    def lfs_per_process(self, val):
        if isinstance(val, int):
            self._lfs_per_process = val
        else:
            raise TypeError(expected_type=int, actual_value=type(val))

    @upload_input_data.setter
    def upload_input_data(self, val):
        if isinstance(val, list):
            self._upload_input_data = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))

    @copy_input_data.setter
    def copy_input_data(self, val):
        if isinstance(val, list):
            self._copy_input_data = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))


    @move_input_data.setter
    def move_input_data(self, val):
        if isinstance(val, list):
            self._move_input_data = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))

    @link_input_data.setter
    def link_input_data(self, val):
        if isinstance(val, list):
            self._link_input_data = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))

    @copy_output_data.setter
    def copy_output_data(self, val):
        if isinstance(val, list):
            self._copy_output_data = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))


    @move_output_data.setter
    def move_output_data(self, val):
        if isinstance(val, list):
            self._move_output_data = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))

    @download_output_data.setter
    def download_output_data(self, val):
        if isinstance(val, list):
            self._download_output_data = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))

    @exit_code.setter
    def exit_code(self, val):
        if isinstance(val, int):
            self._exit_code = val
        else:
            raise TypeError(entity='exit_code',
                            expected_type=int, actual_type=type(val))

    @path.setter
    def path(self, val):
        if isinstance(val, str):
            self._path = val
        else:
            raise TypeError(entity='path', expected_type=str,
                            actual_type=type(val))

    @tag.setter
    def tag(self, val):
        if isinstance(val, str):
            self._tag = val
        else:
            raise TypeError(entity='tag', expected_type=str,
                            actual_type=type(val))

    @parent_stage.setter
    def parent_stage(self, val):
        if isinstance(val, dict):
            self._p_stage = val
        else:
            raise TypeError(expected_type=dict, actual_type=type(val))

    @parent_pipeline.setter
    def parent_pipeline(self, val):
        if isinstance(val, dict):
            self._p_pipeline = val
        else:
            raise TypeError(expected_type=dict, actual_type=type(val))

    # ------------------------------------------------------------------------------------------------------------------
    # Public methods
    # ------------------------------------------------------------------------------------------------------------------

    def to_dict(self):
        """
        Convert current Task into a dictionary

        :return: python dictionary
        """

        task_desc_as_dict = {
            'uid': self._uid,
            'name': self._name,
            'state': self._state,
            'state_history': self._state_history,

            'pre_exec': self._pre_exec,
            'executable': self._executable,
            'arguments': self._arguments,
            'post_exec': self._post_exec,
            'cpu_reqs': self._cpu_reqs,
            'gpu_reqs': self._gpu_reqs,
            'lfs_per_process': self._lfs_per_process,

            'upload_input_data': self._upload_input_data,
            'copy_input_data': self._copy_input_data,
            'link_input_data': self._link_input_data,
            'move_input_data': self._move_input_data,
            'copy_output_data': self._copy_output_data,
            'move_output_data': self._move_output_data,
            'download_output_data': self._download_output_data,

            'exit_code': self._exit_code,
            'path': self._path,
            'tag': self._tag,

            'parent_stage': self._p_stage,
            'parent_pipeline': self._p_pipeline,
        }

        return task_desc_as_dict

    def from_dict(self, d):
        """
        Create a Task from a dictionary. The change is in inplace.

        :argument: python dictionary
        :return: None
        """

        if 'uid' in d:
            if d['uid']:
                self._uid = d['uid']

        if 'name' in d:
            if d['name']:
                self._name = d['name']

        if 'state' in d:
            if isinstance(d['state'], str) or isinstance(d['state'], unicode):
                self._state = d['state']
            else:
                raise TypeError(entity='state', expected_type=str,
                                actual_type=type(d['state']))
        else:
            self._state = states.INITIAL

        if 'state_history' in d:
            if isinstance(d['state_history'], list):
                self._state_history = d['state_history']
            else:
                raise TypeError(entity='state_history', expected_type=list, actual_type=type(
                    d['state_history']))

        if 'pre_exec' in d:
            if isinstance(d['pre_exec'], list):
                self._pre_exec = d['pre_exec']
            else:
                raise TypeError(expected_type=list,
                                actual_type=type(d['pre_exec']))

        if 'executable' in d:
            if isinstance(d['executable'], list):
                self._executable = d['executable']
            else:
                raise TypeError(expected_type=list,
                                actual_type=type(d['executable']))

        if 'arguments' in d:
            if isinstance(d['arguments'], list):
                self._arguments = d['arguments']
            else:
                raise TypeError(expected_type=list,
                                actual_type=type(d['arguments']))

        if 'post_exec' in d:
            if isinstance(d['post_exec'], list):
                self._post_exec = d['post_exec']
            else:
                raise TypeError(expected_type=list,
                                actual_type=type(d['post_exec']))

        if 'cpu_reqs' in d:
            if isinstance(d['cpu_reqs'], dict):
                self._cpu_reqs = d['cpu_reqs']
            else:
                raise TypeError(expected_type=dict,
                                actual_type=type(d['cpu_reqs']))

        if 'gpu_reqs' in d:
            if isinstance(d['gpu_reqs'], dict):
                self._gpu_reqs = d['gpu_reqs']
            else:
                raise TypeError(expected_type=dict,
                                actual_type=type(d['gpu_reqs']))

        if 'lfs_per_process' in d:
            if d['lfs_per_process']:
                if isinstance(d['lfs_per_process'], int):
                    self._lfs_per_process = d['lfs_per_process']
                else:
                    raise TypeError(expected_type=int,
                                    actual_type=type(d['lfs_per_process']))

        if 'upload_input_data' in d:
            if isinstance(d['upload_input_data'], list):
                self._upload_input_data = d['upload_input_data']
            else:
                raise TypeError(expected_type=list,
                                actual_type=type(d['upload_input_data']))

        if 'copy_input_data' in d:
            if isinstance(d['copy_input_data'], list):
                self._copy_input_data = d['copy_input_data']
            else:
                raise TypeError(expected_type=list,
                                actual_type=type(d['copy_input_data']))

        if 'link_input_data' in d:
            if isinstance(d['link_input_data'], list):
                self._link_input_data = d['link_input_data']
            else:
                raise TypeError(expected_type=list,
                                actual_type=type(d['link_input_data']))

        if 'move_input_data' in d:
            if isinstance(d['move_input_data'], list):
                self._move_input_data = d['move_input_data']
            else:
                raise TypeError(expected_type=list,
                                actual_type=type(d['move_input_data']))


        if 'copy_output_data' in d:
            if isinstance(d['copy_output_data'], list):
                self._copy_output_data = d['copy_output_data']
            else:
                raise TypeError(expected_type=list,
                                actual_type=type(d['copy_output_data']))

        if 'move_output_data' in d:
            if isinstance(d['move_output_data'], list):
                self._move_output_data = d['move_output_data']
            else:
                raise TypeError(expected_type=list,
                                actual_type=type(d['move_output_data']))

        if 'download_output_data' in d:
            if isinstance(d['download_output_data'], list):
                self._download_output_data = d['download_output_data']
            else:
                raise TypeError(expected_type=list, actual_type=type(
                    d['download_output_data']))

        if 'exit_code' in d:
            if d['exit_code']:
                if isinstance(d['exit_code'], int):
                    self._exit_code = d['exit_code']
                else:
                    raise TypeError(
                        entity='exit_code', expected_type=int, actual_type=type(d['exit_code']))

        if 'path' in d:
            if d['path']:
                if isinstance(d['path'], str) or isinstance(d['path'], unicode):
                    self._path = d['path']
                else:
                    raise TypeError(entity='path', expected_type=str,
                                    actual_type=type(d['path']))

        if 'tag' in d:
            if d['tag']:
                if isinstance(d['tag'], str) or isinstance(d['tag'], unicode):
                    self._tag = str(d['tag'])
                else:
                    raise TypeError(expected_type=str,
                                    actual_type=type(d['tag']))

        if 'parent_stage' in d:
            if isinstance(d['parent_stage'], dict):
                self._p_stage = d['parent_stage']
            else:
                raise TypeError(
                    entity='parent_stage', expected_type=dict, actual_type=type(d['parent_stage']))

        if 'parent_pipeline' in d:
            if isinstance(d['parent_pipeline'], dict):
                self._p_pipeline = d['parent_pipeline']
            else:
                raise TypeError(entity='parent_pipeline', expected_type=dict, actual_type=type(
                    d['parent_pipeline']))

    # ------------------------------------------------------------------------------------------------------------------
    # Private methods
    # ------------------------------------------------------------------------------------------------------------------

    def _assign_uid(self, sid):
        """
        Purpose: Assign a uid to the current object based on the sid passed
        """
        self._uid = ru.generate_id(
            'task.%(item_counter)04d', ru.ID_CUSTOM, namespace=sid)

    def _validate(self):
        """
        Purpose: Validate that the state of the task is 'DESCRIBED' and that an executable has been specified for the
        task.
        """

        if self._state is not states.INITIAL:
            raise ValueError(obj=self._uid,
                             attribute='state',
                             expected_value=states.INITIAL,
                             actual_value=self._state)

        if not self._executable:
            raise MissingError(obj=self._uid,
                               missing_attribute='executable')
    # ------------------------------------------------------------------------------------------------------------------



__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__ = "Vivek Balasubramanian <vivek.balasubramaniana@rutgers.edu>"
__license__ = "MIT"

import radical.utils as ru
from radical.entk.exceptions import *
import threading
from multiprocessing import Process, Event
import Queue
import time
import json
import pika
import traceback
import os
import uuid
from resource_manager import Base_ResourceManager


class Base_TaskManager(object):

    """
    A Task Manager takes the responsibility of dispatching tasks it receives from a pending_queue for execution on to 
    the available resources using a runtime system. Once the tasks have completed execution, they are pushed on to 
    the completed_queue for other components of EnTK to process.

    :arguments:
        :pending_queue: List of queue(s) with tasks ready to be executed. Currently, only one queue.
        :completed_queue: List of queue(s) with tasks that have finished execution. Currently, only one queue.
        :rmgr: ResourceManager object to be used to access the Pilot where the tasks can be submitted
        :mq_hostname: Name of the host where RabbitMQ is running
        :port: port at which rabbitMQ can be accessed

    Currently, EnTK is configured to work with one pending queue and one completed queue. In the future, the number of 
    queues can be varied for different throughput requirements at the cost of additional Memory and CPU consumption.
    """

    def __init__(self,
                 sid,
                 pending_queue,
                 completed_queue,
                 rmgr,
                 mq_hostname,
                 port,
                 rts):

        if isinstance(sid, str):
            self._sid = sid
        else:
            raise TypeError(expected_type=str, actual_type=type(sid))

        if isinstance(pending_queue, list):
            self._pending_queue = pending_queue
        else:
            raise TypeError(expected_type=str, actual_type=type(pending_queue))

        if isinstance(completed_queue, list):
            self._completed_queue = completed_queue
        else:
            raise TypeError(expected_type=str, actual_type=type(completed_queue))

        if isinstance(mq_hostname, str):
            self._mq_hostname = mq_hostname
        else:
            raise TypeError(expected_type=str, actual_type=type(mq_hostname))

        if isinstance(port, int):
            self._port = port
        else:
            raise TypeError(expected_type=int, actual_type=type(port))

        if isinstance(rmgr, Base_ResourceManager):
            self._rmgr = rmgr
        else:
            raise TypeError(expected_type=ResourceManager, actual_type=type(rmgr))

        self._rts = rts

        # Utility parameters
        self._uid = ru.generate_id('task_manager.%(item_counter)04d', ru.ID_CUSTOM, namespace=self._sid)
        self._path = os.getcwd() + '/' + self._sid
        self._logger = ru.Logger('radical.entk.%s' %
                                 self._uid, path=self._path, targets=['2', '.'])
        self._prof = ru.Profiler(name='radical.entk.%s' % self._uid + '-obj', path=self._path)

        # Thread should run till terminate condtion is encountered
        mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=mq_hostname, port=port))

        self._hb_request_q = '%s-hb-request' % self._sid
        self._hb_response_q = '%s-hb-response' % self._sid

        mq_channel = mq_connection.channel()

        # To respond to heartbeat - get request from rpc_queue
        mq_channel.queue_delete(queue=self._hb_response_q)
        mq_channel.queue_declare(queue=self._hb_response_q)

        # To respond to heartbeat - get request from rpc_queue
        mq_channel.queue_delete(queue=self._hb_request_q)
        mq_channel.queue_declare(queue=self._hb_request_q)

        self._tmgr_process = None
        self._hb_thread = None
        self._hb_interval = int(os.getenv('ENTK_HB_INTERVAL', 30))

        mq_connection.close()

    # ------------------------------------------------------------------------------------------------------------------
    # Private Methods
    # ------------------------------------------------------------------------------------------------------------------

    def _heartbeat(self):
        """
        **Purpose**: Method to be executed in the heartbeat thread. This method sends a 'request' to the
        heartbeat-req queue. It expects a 'response' message from the 'heartbeart-res' queue within 10 seconds. This
        message should contain the same correlation id. If no message if received in 10 seconds, the tmgr is assumed
        dead. The end_manager() is called to cleanly terminate tmgr process and the heartbeat thread is also 
        terminated.

        **Details**: The AppManager can re-invoke both if the execution is still not complete.
        """

        try:

            self._prof.prof('heartbeat thread started', uid=self._uid)

            mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=self._mq_hostname, port=self._port))
            mq_channel = mq_connection.channel()

            response = True
            while (response and (not self._hb_terminate.is_set())):
                response = False
                corr_id = str(uuid.uuid4())

                # Heartbeat request signal sent to task manager via rpc-queue
                mq_channel.basic_publish(exchange='',
                                         routing_key=self._hb_request_q,
                                         properties=pika.BasicProperties(
                                             reply_to=self._hb_response_q,
                                             correlation_id=corr_id),
                                         body='request')
                self._logger.info('Sent heartbeat request')

                # mq_connection.close()

                # Sleep for hb_interval and then check if tmgr responded
                mq_connection.sleep(self._hb_interval)

                # mq_connection = pika.BlockingConnection(
                #     pika.ConnectionParameters(host=self._mq_hostname, port=self._port))
                # mq_channel = mq_connection.channel()

                method_frame, props, body = mq_channel.basic_get(queue=self._hb_response_q)

                if body:
                    if corr_id == props.correlation_id:
                        self._logger.info('Received heartbeat response')
                        response = True

                        mq_channel.basic_ack(delivery_tag=method_frame.delivery_tag)

                # Appease pika cos it thinks the connection is dead
                # mq_connection.close()

        except KeyboardInterrupt:
            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to cancel tmgr process gracefully...')
            raise KeyboardInterrupt

        except Exception as ex:
            self._logger.exception('Heartbeat failed with error: %s' % ex)
            raise

        finally:

            try:
                mq_connection.close()
            except:
                self._logger.warning('mq_connection not created')

            self._prof.prof('terminating heartbeat thread', uid=self._uid)

    def _tmgr(self, uid, rmgr, logger, mq_hostname, port, pending_queue, completed_queue):
        """
        **Purpose**: Method to be run by the tmgr process. This method receives a Task from the pending_queue
        and submits it to the RTS. At all state transititons, they are synced (blocking) with the AppManager
        in the master process.

        In addition, the tmgr also receives heartbeat 'request' msgs from the heartbeat-req queue. It responds with a
        'response' message to the 'heartbeart-res' queue.

        **Details**: The AppManager can re-invoke the tmgr process with this function if the execution of the workflow is 
        still incomplete. There is also population of a dictionary, placeholder_dict, which stores the path of each of
        the tasks on the remote machine. 
        """

        raise NotImplementedError('_tmgr() method ' +
                                  'not implemented in TaskManager for %s' % self._rts)

    # ------------------------------------------------------------------------------------------------------------------
    # Public Methods
    # ------------------------------------------------------------------------------------------------------------------

    def start_heartbeat(self):
        """
        **Purpose**: Method to start the heartbeat thread. The heartbeat function
        is not to be accessed directly. The function is started in a separate
        thread using this method.
        """

        if not self._hb_thread:

            try:

                self._logger.info('Starting heartbeat thread')
                self._prof.prof('creating heartbeat thread', uid=self._uid)
                self._hb_terminate = threading.Event()
                self._hb_thread = threading.Thread(target=self._heartbeat, name='heartbeat')

                self._prof.prof('starting heartbeat thread', uid=self._uid)
                self._hb_thread.start()

                return True

            except Exception, ex:

                self._logger.error('Heartbeat not started, error: %s' % ex)
                self.terminate_heartbeat()
                raise

        else:
            self._logger.warn('Heartbeat thread already running, but attempted to restart!')

    def terminate_heartbeat(self):
        """
        **Purpose**: Method to terminate the heartbeat thread. This method is 
        blocking as it waits for the heartbeat thread to terminate (aka join).

        This is the last method that is executed from the TaskManager and
        hence closes the profiler.
        """

        try:

            if self._hb_thread:

                self._hb_terminate.set()

                if self.check_heartbeat():
                    self._hb_thread.join()

                self._hb_thread = None

                self._logger.info('Hearbeat thread terminated')

                self._prof.prof('heartbeat thread terminated', uid=self._uid)

                # We close in the heartbeat because it ends after the tmgr process
                self._prof.close()

        except Exception, ex:
            self._logger.error('Could not terminate heartbeat thread')
            raise

        finally:

            if not (self.check_heartbeat() or self.check_manager()):

                mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=self._mq_hostname, port=self._port))
                mq_channel = mq_connection.channel()

                # To respond to heartbeat - get request from rpc_queue
                mq_channel.queue_delete(queue=self._hb_response_q)
                mq_channel.queue_delete(queue=self._hb_request_q)

                mq_connection.close()

    def start_manager(self):
        """
        **Purpose**: Method to start the tmgr process. The tmgr function
        is not to be accessed directly. The function is started in a separate
        thread using this method.
        """

        raise NotImplementedError('start_manager() method ' +
                                  'not implemented in TaskManager for %s' % self._rts)

    def terminate_manager(self):
        """
        **Purpose**: Method to terminate the tmgr process. This method is 
        blocking as it waits for the tmgr process to terminate (aka join).
        """

        try:

            if self._tmgr_process:

                if not self._tmgr_terminate.is_set():
                    self._tmgr_terminate.set()

                if self.check_manager():
                    self._tmgr_process.join()

                self._tmgr_process = None
                self._logger.info('Task manager process closed')

                self._prof.prof('tmgr process terminated', uid=self._uid)

        except Exception, ex:
            self._logger.error('Could not terminate task manager process')
            raise



    def check_heartbeat(self):
        """
        **Purpose**: Check if the heartbeat thread is alive and running
        """
        if self._hb_thread:
            return self._hb_thread.is_alive()
        else:
            return None

    def check_manager(self):
        """
        **Purpose**: Check if the tmgr process is alive and running
        """

        if self._tmgr_process:
            return self._tmgr_process.is_alive()
        else:
            return None

    # ------------------------------------------------------------------------------------------------------------------



import uuid
import json
import pika


def sync_with_master(obj, obj_type, channel, queue, logger, local_prof):

    object_as_dict = {'object': obj.to_dict()}
    if obj_type == 'Task':
        object_as_dict['type'] = 'Task'

    elif obj_type == 'Stage':
        object_as_dict['type'] = 'Stage'

    elif obj_type == 'Pipeline':
        object_as_dict['type'] = 'Pipeline'

    corr_id = str(uuid.uuid4())

    logger.debug('Attempting to sync %s with state %s with AppManager' % (obj.uid, obj.state))
    channel.basic_publish(exchange='',
                          routing_key=queue,
                          body=json.dumps(object_as_dict),
                          properties=pika.BasicProperties(correlation_id=corr_id)
                          )

    if obj_type == 'Task':
        local_prof.prof('publishing obj with state %s for sync' % obj.state, uid=obj.uid, msg=obj.parent_stage['uid'])
    elif obj_type == 'Stage':
        local_prof.prof('publishing obj with state %s for sync' %
                        obj.state, uid=obj.uid, msg=obj.parent_pipeline['uid'])
    else:
        local_prof.prof('publishing obj with state %s for sync' % obj.state, uid=obj.uid)

    sid = '-'.join(queue.split('-')[:-3])
    qname = queue.split('-')[-3:]
    qname.reverse()
    reply_queue = '-'.join(qname)
    reply_queue = sid + '-' + reply_queue

    while True:
        #self._logger.info('waiting for ack')

        method_frame, props, body = channel.basic_get(queue=reply_queue)

        if body:
            if corr_id == props.correlation_id:

                # print 'acknowledged: ', obj.uid, obj.state
                if obj_type == 'Task':
                    local_prof.prof('obj with state %s synchronized' %
                                    obj.state, uid=obj.uid, msg=obj.parent_stage['uid'])
                elif obj_type == 'Stage':
                    local_prof.prof('obj with state %s synchronized' %
                                    obj.state, uid=obj.uid, msg=obj.parent_pipeline['uid'])
                else:
                    local_prof.prof('obj with state %s synchronized' % obj.state, uid=obj.uid)

                logger.debug('%s with state %s synced with AppManager' % (obj.uid, obj.state))

                channel.basic_ack(delivery_tag=method_frame.delivery_tag)

                break



from radical.entk import Pipeline, Stage, Task, AppManager
import os

# ------------------------------------------------------------------------------
# Set default verbosity
if os.environ.get('RADICAL_ENTK_VERBOSE') == None:
    os.environ['RADICAL_ENTK_REPORT'] = 'True'


# Description of how the RabbitMQ process is accessible
# No need to change/set any variables if you installed RabbitMQ has a system
# process. If you are running RabbitMQ under a docker container or another
# VM, set "RMQ_HOSTNAME" and "RMQ_PORT" in the session where you are running
# this script.
hostname = os.environ.get('RMQ_HOSTNAME', 'localhost')
port = os.environ.get('RMQ_PORT', 5672)


def generate_pipeline():

    # Create a Pipeline object
    p = Pipeline()
    p.name = 'p1'

    # Create a Stage object
    s1 = Stage()
    s1.name = 's1'

    # Create a Task object which creates a file named 'output.txt' of size 1 MB
    t1 = Task()
    t1.name = 't1'
    t1.executable = ['/bin/bash']
    t1.arguments = ['-l', '-c', 'base64 /dev/urandom | head -c 1000000 > output.txt']

    # Add the Task to the Stage
    s1.add_tasks(t1)

    # Add Stage to the Pipeline
    p.add_stages(s1)

    # Create another Stage object to hold character count tasks
    s2 = Stage()
    s2.name = 's2'
    s2_task_uids = []

    for cnt in range(10):

        # Create a Task object
        t2 = Task()
        t2.name = 't%s' % (cnt + 1)
        t2.executable = ['/bin/bash']
        t2.arguments = ['-l', '-c', 'grep -o . output.txt | sort | uniq -c > ccount.txt']
        # Copy data from the task in the first stage to the current task's location
        t2.copy_input_data = ['$Pipline_%s_Stage_%s_Task_%s/output.txt' % (p.name, s1.name, t1.name)]

        # Add the Task to the Stage
        s2.add_tasks(t2)
        s2_task_uids.append(t2.name)

    # Add Stage to the Pipeline
    p.add_stages(s2)

    # Create another Stage object to hold checksum tasks
    s3 = Stage()
    s3.name = 's3'

    for cnt in range(10):

        # Create a Task object
        t3 = Task()
        t3.name = 't%s' % (cnt + 1)
        t3.executable = ['/bin/bash']
        t3.arguments = ['-l', '-c', 'sha1sum ccount.txt > chksum.txt']
        # Copy data from the task in the first stage to the current task's location
        t3.copy_input_data = ['$Pipline_%s_Stage_%s_Task_%s/ccount.txt' % (p.name, s2.name, s2_task_uids[cnt])]
        # Download the output of the current task to the current location
        t3.download_output_data = ['chksum.txt > chksum_%s.txt' % cnt]

        # Add the Task to the Stage
        s3.add_tasks(t3)

    # Add Stage to the Pipeline
    p.add_stages(s3)

    return p

if __name__ == '__main__':

    # Create Application Manager
    appman = AppManager(hostname=hostname, port=port)

    # Create a dictionary describe four mandatory keys:
    # resource, walltime, and cpus
    # resource is 'local.localhost' to execute locally
    res_dict = {

        'resource': 'local.localhost',
        'walltime': 10,
        'cpus': 1
    }

    # Assign resource request description to the Application Manager
    appman.resource_desc = res_dict

    # Assign the workflow as a set or list of Pipelines to the Application Manager
    # Note: The list order is not guaranteed to be preserved
    appman.workflow = set([generate_pipeline()])

    # Run the Application Manager
    appman.run()



__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__ = "Vivek Balasubramanian <vivek.balasubramanian@rutgers.edu>"
__license__ = "MIT"

import radical.utils as ru
from radical.entk.exceptions import *
import threading
from multiprocessing import Process, Event
import Queue
from radical.entk import states, Task
from radical.entk.utils.init_transition import transition
import time
import json
import pika
import traceback
import os
import radical.pilot as rp
from task_processor import create_cud_from_task, create_task_from_cu
import uuid
from ..base.task_manager import Base_TaskManager


class TaskManager(Base_TaskManager):

    """
    A Task Manager takes the responsibility of dispatching tasks it receives from a queue for execution on to 
    the available resources using a runtime system. In this case, the runtime system being used RADICAL Pilot. Once 
    the tasks have completed execution, they are pushed on to another queue for other components of EnTK to access.


    :arguments:
        :pending_queue: List of queue(s) with tasks ready to be executed. Currently, only one queue.
        :completed_queue: List of queue(s) with tasks that have finished execution. Currently, only one queue.
        :rmgr: ResourceManager object to be used to access the Pilot where the tasks can be submitted
        :mq_hostname: Name of the host where RabbitMQ is running
        :port: port at which rabbitMQ can be accessed

    Currently, EnTK is configured to work with one pending queue and one completed queue. In the future, the number of 
    queues can be varied for different throughput requirements at the cost of additional Memory and CPU consumption.
    """

    def __init__(self, sid, pending_queue, completed_queue,
                rmgr, mq_hostname, port):

        super(TaskManager, self).__init__(sid,
                                        pending_queue,
                                        completed_queue,
                                        rmgr,
                                        mq_hostname,
                                        port,
                                        rts='radical.pilot')

        self._umgr = None
        self._rmq_ping_interval = os.getenv('RMQ_PING_INTERVAL', 10)
        self._logger.info('Created task manager object: %s' % self._uid)
        self._prof.prof('tmgr obj created', uid=self._uid)

    # ------------------------------------------------------------------------------------------------------------------
    # Private Methods
    # ------------------------------------------------------------------------------------------------------------------

    def _tmgr(self, uid, umgr, rmgr, logger, mq_hostname, port, pending_queue, completed_queue):
        """
        **Purpose**: Method to be run by the tmgr process. This method receives a Task from the pending_queue
        and submits it to the RTS. Currently, it also converts Tasks into CUDs and CUs into (partially described) Tasks.
        This conversion is necessary since the current RTS is RADICAL Pilot. Once Tasks are recovered from a CU, they
        are then pushed to the completed_queue. At all state transititons, they are synced (blocking) with the AppManager
        in the master process.

        In addition the tmgr also receives heartbeat 'request' msgs from the heartbeat-request queue. It responds with a
        'response' message to the heartbeart-response queue.

        **Details**: The AppManager can re-invoke the tmgr process with this function if the execution of the workflow is 
        still incomplete. There is also population of a dictionary, placeholder_dict, which stores the path of each of
        the tasks on the remote machine. 
        """

        try:

            local_prof = ru.Profiler(name='radical.entk.%s' % self._uid + '-proc', path=self._path)

            local_prof.prof('tmgr process started', uid=self._uid)
            logger.info('Task Manager process started')

            placeholder_dict = dict()

            def load_placeholder(task, rts_uid):

                parent_pipeline = str(task.parent_pipeline['name'])
                parent_stage = str(task.parent_stage['name'])

                if parent_pipeline not in placeholder_dict:
                    placeholder_dict[parent_pipeline] = dict()

                if parent_stage not in placeholder_dict[parent_pipeline]:
                    placeholder_dict[parent_pipeline][parent_stage] = dict()

                if None not in [parent_pipeline, parent_stage, task.name]:
                    placeholder_dict[parent_pipeline][parent_stage][str(task.name)] = {'path': str(task.path),
                                                                                        'rts_uid': rts_uid}

            def heartbeat_response(mq_channel):

                try:

                    # Get request from heartbeat-req for heartbeat response
                    hb_method_frame, hb_props, hb_body = mq_channel.basic_get(queue=self._hb_request_q)

                    if hb_body:

                        logger.info('Received heartbeat request')

                        mq_channel.basic_publish(exchange='',
                                                routing_key=self._hb_response_q,
                                                properties=pika.BasicProperties(correlation_id=hb_props.correlation_id),
                                                body='response')

                        logger.info('Sent heartbeat response')
                        mq_channel.basic_ack(delivery_tag=hb_method_frame.delivery_tag)

                except Exception, ex:
                    logger.exception('Failed to respond to heartbeat request, error: %s' % ex)
                    raise

            def unit_state_cb(unit, state):

                try:

                    logger.debug('Unit %s in state %s' % (unit.uid, unit.state))

                    if unit.state in rp.FINAL:

                        # Acquire a connection+channel to the rmq server
                        mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=mq_hostname, port=port))
                        mq_channel = mq_connection.channel()

                        task = None
                        task = create_task_from_cu(unit, local_prof)

                        transition(obj=task,
                                obj_type='Task',
                                new_state=states.COMPLETED,
                                channel=mq_channel,
                                queue='%s-cb-to-sync' % self._sid,
                                profiler=local_prof,
                                logger=logger)

                        load_placeholder(task, unit.uid)

                        task_as_dict = json.dumps(task.to_dict())

                        mq_channel.basic_publish(exchange='',
                                                routing_key='%s-completedq-1' % self._sid,
                                                body=task_as_dict
                                                # properties=pika.BasicProperties(
                                                # make message persistent
                                                #    delivery_mode = 2,
                                                #)
                                                )

                        logger.info('Pushed task %s with state %s to completed queue %s' % (task.uid, task.state,
                                                                                            completed_queue[0]))

                        mq_connection.close()

                except KeyboardInterrupt:
                    self._logger.exception('Execution interrupted by user (you probably hit Ctrl+C), ' +
                                            'trying to exit callback thread gracefully...')

                    raise KeyboardInterrupt

                except Exception, ex:
                    self._logger.exception('Error in RP callback thread: %s' % ex)

            if not umgr:
                umgr = rp.UnitManager(session=rmgr._session)
                umgr.add_pilots(rmgr.pilot)
                umgr.register_callback(unit_state_cb)

            # Acquire a connection+channel to the rmq server
            mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=mq_hostname, port=port))
            mq_channel = mq_connection.channel()

            # Make sure the heartbeat response queue is empty
            mq_channel.queue_delete(queue=self._hb_response_q)
            mq_channel.queue_declare(queue=self._hb_response_q)

            local_prof.prof('tmgr infrastructure setup done', uid=uid)

            last = time.time()
            while not self._tmgr_terminate.is_set():

                try:

                    # Get tasks from the pending queue
                    method_frame, header_frame, body = mq_channel.basic_get(queue=pending_queue[0])

                    if body:

                        body = json.loads(body)
                        bulk_tasks = list()
                        bulk_cuds = list()

                        for task in body:
                            t = Task()
                            t.from_dict(task)
                            bulk_tasks.append(t)
                            bulk_cuds.append(create_cud_from_task(t, placeholder_dict, local_prof))

                            transition(obj=t,
                                    obj_type='Task',
                                    new_state=states.SUBMITTING,
                                    channel=mq_channel,
                                    queue='%s-tmgr-to-sync' % self._sid,
                                    profiler=local_prof,
                                    logger=self._logger)

                        # To accommodate long cud creation times
                        heartbeat_response(mq_channel)

                        now = time.time()
                        if now - last >= self._rmq_ping_interval:
                            mq_connection.process_data_events()
                            last = now

                        umgr.submit_units(bulk_cuds)

                        for task in bulk_tasks:

                            transition(obj=task,
                                    obj_type='Task',
                                    new_state=states.SUBMITTED,
                                    channel=mq_channel,
                                    queue='%s-tmgr-to-sync' % self._sid,
                                    profiler=local_prof,
                                    logger=self._logger)
                            self._logger.info('Task %s submitted to RTS' % (task.uid))


                        # To accommodate long cud submission times
                        heartbeat_response(mq_channel)

                        # Appease pika cos it thinks the connection is dead
                        now = time.time()
                        if now - last >= self._rmq_ping_interval:
                            mq_connection.process_data_events()
                            last = now

                        mq_channel.basic_ack(delivery_tag=method_frame.delivery_tag)

                    heartbeat_response(mq_channel)

                except Exception, ex:
                    logger.exception('Error in task execution: %s' % ex)
                    raise

            local_prof.prof('terminating tmgr process', uid=uid)
            mq_connection.close()
            local_prof.close()

        except KeyboardInterrupt:

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                                'trying to cancel tmgr process gracefully...')
            raise KeyboardInterrupt

        except Exception, ex:

            print traceback.format_exc()
            raise EnTKError(ex)

    # ------------------------------------------------------------------------------------------------------------------
    # Public Methods
    # ------------------------------------------------------------------------------------------------------------------

    def start_manager(self):
        """
        **Purpose**: Method to start the tmgr process. The tmgr function
        is not to be accessed directly. The function is started in a separate
        thread using this method.
        """

        if not self._tmgr_process:

            try:

                self._prof.prof('creating tmgr process', uid=self._uid)
                self._tmgr_terminate = Event()

                self._tmgr_process = Process(target=self._tmgr,
                                            name='task-manager',
                                            args=(
                                                self._uid,
                                                self._umgr,
                                                self._rmgr,
                                                self._logger,
                                                self._mq_hostname,
                                                self._port,
                                                self._pending_queue,
                                                self._completed_queue)
                                            )

                self._logger.info('Starting task manager process')
                self._prof.prof('starting tmgr process', uid=self._uid)
                self._tmgr_process.start()

                return True

            except Exception, ex:

                self._logger.error('Task manager not started, error: %s' % ex)
                self.terminate_manager()
                raise

        else:
            self._logger.warn('tmgr process already running, but attempted to restart!')
    # ------------------------------------------------------------------------------------------------------------------



__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__ = "Vivek Balasubramanian <vivek.balasubramaniana@rutgers.edu>"
__license__ = "MIT"

import radical.utils as ru
from radical.entk.exceptions import *
from multiprocessing import Process, Event
from radical.entk import states, Pipeline, Task
from radical.entk.utils.init_transition import transition
import time
from time import sleep
import json
import threading
import pika
import traceback
import os
import uuid
import time


class WFprocessor(object):

    """
    An WFprocessor (workflow processor) takes the responsibility of dispatching tasks from the various pipelines of the
    workflow according to their relative order to the TaskManager. All state updates are communicated to the AppManager.
    The WFprocessor also retrieves completed tasks from the TaskManager and updates states of PST accordingly.

    :Arguments:
        :sid: (str) session id to be used by the profiler and loggers
        :workflow: (set) COPY of the entire workflow existing in the AppManager
        :pending_queue: (list) queues to hold pending tasks
        :completed_queue: (list) queues to hold completed tasks
        :mq_hostname: (str) hostname where the RabbitMQ is alive
        :port: (int) port at which RabbitMQ can be accessed
        :resubmit_failed: (bool) True if failed tasks need to be resubmitted automatically
    """

    def __init__(self,
                 sid,
                 workflow,
                 pending_queue,
                 completed_queue,
                 mq_hostname,
                 port,
                 resubmit_failed):

        # Mandatory arguments
        self._sid = sid
        self._pending_queue = pending_queue
        self._completed_queue = completed_queue
        self._mq_hostname = mq_hostname
        self._port = port
        self._resubmit_failed = resubmit_failed

        # Assign validated workflow
        self._workflow = workflow

        # Create logger and profiler at their specific locations using the sid
        self._uid = ru.generate_id('wfprocessor.%(item_counter)04d', ru.ID_CUSTOM, namespace=self._sid)
        self._path = os.getcwd() + '/' + self._sid
        self._logger = ru.Logger('radical.entk.%s' %
                                 self._uid, path=self._path, targets=['2', '.'])
        self._prof = ru.Profiler(name='radical.entk.%s' % self._uid + '-obj', path=self._path)

        self._prof.prof('create wfp obj', uid=self._uid)

        # Defaults
        self._wfp_process = None
        self._rmq_ping_interval = os.getenv('RMQ_PING_INTERVAL', 10)

        self._logger.info('Created WFProcessor object: %s' % self._uid)
        self._prof.prof('wfp obj created', uid=self._uid)

    # ------------------------------------------------------------------------------------------------------------------
    # Getter
    # ------------------------------------------------------------------------------------------------------------------
    @property
    def workflow(self):
        return self._workflow

    # ------------------------------------------------------------------------------------------------------------------
    # Private Methods
    # ------------------------------------------------------------------------------------------------------------------

    def _initialize_workflow(self):
        """
        **Purpose**: Initialize the PST of the workflow with a uid and type checks
        """

        try:

            self._prof.prof('initializing workflow', uid=self._uid)

            for p in self._workflow:
                p._assign_uid(self._sid)

            self._prof.prof('workflow initialized', uid=self._uid)

        except Exception, ex:
            self._logger.exception('Fatal error while initializing workflow: %s' % ex)
            raise

    def _enqueue(self, local_prof):
        """
        **Purpose**: This is the function that is run in the enqueue thread. This function extracts Tasks from the
        copy of workflow that exists in the WFprocessor object and pushes them to the queues in the pending_q list.
        Since this thread works on the copy of the workflow, every state update to the Task, Stage and Pipeline is
        communicated back to the AppManager (master process) via the 'sync_with_master' function that has dedicated
        queues to communicate with the master.

        Details: Termination condition of this thread is set by the wfp process.
        """

        try:

            local_prof.prof('enqueue-thread started', uid=self._uid)
            self._logger.info('enqueue-thread started')

            # Acquire a connection+channel to the rmq server
            mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=self._mq_hostname, port=self._port))
            mq_channel = mq_connection.channel()

            last = time.time()
            while not self._enqueue_thread_terminate.is_set():

                '''
                We iterate through all pipelines to collect tasks from
                stages that are pending scheduling. Once collected, these tasks
                will be communicated to the tmgr in bulk.
                '''

                workload = []
                scheduled_stages = []

                for pipe in self._workflow:

                    with pipe.lock:

                        if not pipe.completed:

                            # Test if the pipeline is already in the final state
                            if pipe.state in states.FINAL:
                                continue

                            elif pipe.state == states.INITIAL:

                                # Set state of pipeline to SCHEDULING if it is in INITIAL
                                transition(obj=pipe,
                                           obj_type='Pipeline',
                                           new_state=states.SCHEDULING,
                                           channel=mq_channel,
                                           queue='%s-enq-to-sync' % self._sid,
                                           profiler=local_prof,
                                           logger=self._logger)

                            executable_stage = pipe.stages[pipe.current_stage - 1]

                            if not executable_stage.uid:
                                executable_stage.parent_pipeline['uid'] = pipe.uid
                                executable_stage.parent_pipeline['name'] = pipe.name
                                executable_stage._assign_uid(self._sid)

                            if executable_stage.state in [states.INITIAL, states.SCHEDULED]:

                                if executable_stage.state == states.INITIAL:

                                    transition(obj=executable_stage,
                                               obj_type='Stage',
                                               new_state=states.SCHEDULING,
                                               channel=mq_channel,
                                               queue='%s-enq-to-sync' % self._sid,
                                               profiler=local_prof,
                                               logger=self._logger)

                                executable_tasks = executable_stage.tasks

                                for executable_task in executable_tasks:

                                    if (executable_task.state == states.INITIAL)or \
                                            ((executable_task.state == states.FAILED)and(self._resubmit_failed)):

                                        # Set state of Tasks in current Stage to SCHEDULING
                                        transition(obj=executable_task,
                                                   obj_type='Task',
                                                   new_state=states.SCHEDULING,
                                                   channel=mq_channel,
                                                   queue='%s-enq-to-sync' % self._sid,
                                                   profiler=local_prof,
                                                   logger=self._logger)

                                        # task_as_dict = json.dumps(executable_task.to_dict())
                                        workload.append(executable_task)

                                        if executable_stage not in scheduled_stages:
                                            scheduled_stages.append(executable_stage)

                if workload:

                    # Put the task on one of the pending_queues
                    workload_as_dict = list()
                    for task in workload:
                        workload_as_dict.append(task.to_dict())
                    workload_as_dict = json.dumps(workload_as_dict)

                    mq_channel.basic_publish(exchange='',
                                             routing_key=self._pending_queue[0],
                                             body=workload_as_dict
                                             # properties=pika.BasicProperties(
                                             # make message persistent
                                             # delivery_mode = 2)
                                             )

                    for task in workload:

                        # Set state of Tasks in current Stage to SCHEDULED
                        transition(obj=task,
                                   obj_type='Task',
                                   new_state=states.SCHEDULED,
                                   channel=mq_channel,
                                   queue='%s-enq-to-sync' % self._sid,
                                   profiler=local_prof,
                                   logger=self._logger)

                        self._logger.debug('Task %s published to pending queue' % task.uid)

                if scheduled_stages:
                    for executable_stage in scheduled_stages:

                        transition(obj=executable_stage,
                                   obj_type='Stage',
                                   new_state=states.SCHEDULED,
                                   channel=mq_channel,
                                   queue='%s-enq-to-sync' % self._sid,
                                   profiler=local_prof,
                                   logger=self._logger)

                # Appease pika cos it thinks the connection is dead
                now = time.time()
                if now - last >= self._rmq_ping_interval:
                    mq_connection.process_data_events()
                    last = now

            self._logger.info('Enqueue thread terminated')
            mq_connection.close()

            local_prof.prof('terminating enqueue-thread', uid=self._uid)

        except KeyboardInterrupt:

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to cancel enqueuer thread gracefully...')

            mq_connection.close()

            raise KeyboardInterrupt

        except Exception, ex:

            self._logger.exception('Error in enqueue-thread: %s' % ex)
            try:
                mq_connection.close()
            except:
                self._logger.warning('mq_connection not created')

            raise

    def _dequeue(self, local_prof):
        """
        **Purpose**: This is the function that is run in the dequeue thread. This function extracts Tasks from the
        completed queus and updates the copy of workflow that exists in the WFprocessor object.
        Since this thread works on the copy of the workflow, every state update to the Task, Stage and Pipeline is
        communicated back to the AppManager (master process) via the 'sync_with_master' function that has dedicated
        queues to communicate with the master.

        Details: Termination condition of this thread is set by the wfp process.
        """

        try:

            local_prof.prof('dequeue-thread started', uid=self._uid)
            self._logger.info('Dequeue thread started')

            mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=self._mq_hostname, port=self._port))
            mq_channel = mq_connection.channel()

            last = time.time()

            while not self._dequeue_thread_terminate.is_set():

                try:

                    method_frame, header_frame, body = mq_channel.basic_get(queue=self._completed_queue[0])

                    if body:

                        # Get task from the message
                        completed_task = Task()
                        completed_task.from_dict(json.loads(body))
                        self._logger.info('Got finished task %s from queue' % (completed_task.uid))

                        transition(obj=completed_task,
                                   obj_type='Task',
                                   new_state=states.DEQUEUEING,
                                   channel=mq_channel,
                                   queue='%s-deq-to-sync' % self._sid,
                                   profiler=local_prof,
                                   logger=self._logger)

                        # Traverse the entire workflow to find out the correct Task
                        for pipe in self._workflow:

                            with pipe.lock:

                                if not pipe.completed:

                                    if completed_task.parent_pipeline['uid'] == pipe.uid:

                                        self._logger.debug('Found parent pipeline: %s' % pipe.uid)

                                        for stage in pipe.stages:

                                            if completed_task.parent_stage['uid'] == stage.uid:
                                                self._logger.debug('Found parent stage: %s' % (stage.uid))

                                                transition(obj=completed_task,
                                                           obj_type='Task',
                                                           new_state=states.DEQUEUED,
                                                           channel=mq_channel,
                                                           queue='%s-deq-to-sync' % self._sid,
                                                           profiler=local_prof,
                                                           logger=self._logger)

                                                if completed_task.exit_code:
                                                    completed_task.state = states.FAILED
                                                else:
                                                    completed_task.state = states.DONE

                                                for task in stage.tasks:

                                                    if task.uid == completed_task.uid:
                                                        task.state = str(completed_task.state)

                                                        if (task.state == states.FAILED) and (self._resubmit_failed):
                                                            task.state = states.INITIAL

                                                        transition(obj=task,
                                                                   obj_type='Task',
                                                                   new_state=task.state,
                                                                   channel=mq_channel,
                                                                   queue='%s-deq-to-sync' % self._sid,
                                                                   profiler=local_prof,
                                                                   logger=self._logger)

                                                        if stage._check_stage_complete():

                                                            transition(obj=stage,
                                                                       obj_type='Stage',
                                                                       new_state=states.DONE,
                                                                       channel=mq_channel,
                                                                       queue='%s-deq-to-sync' % self._sid,
                                                                       profiler=local_prof,
                                                                       logger=self._logger)

                                                            # Check if Stage has a post-exec that needs to be
                                                            # executed

                                                            if stage.post_exec['condition']:

                                                                try:

                                                                    self._logger.info(
                                                                        'Executing post-exec for stage %s' % stage.uid)
                                                                    self._prof.prof('Adap: executing post-exec', 
                                                                        uid=self._uid)

                                                                    func_condition = stage.post_exec['condition']
                                                                    func_on_true = stage.post_exec['on_true']
                                                                    func_on_false = stage.post_exec['on_false']

                                                                    if func_condition():
                                                                        func_on_true()
                                                                    else:
                                                                        func_on_false()

                                                                    self._logger.info(
                                                                        'Post-exec executed for stage %s' % stage.uid)
                                                                    self._prof.prof('Adap: post-exec executed', uid=self._uid)

                                                                except Exception, ex:
                                                                    self._logger.exception(
                                                                        'Execution failed in post_exec of stage %s' % stage.uid)
                                                                    raise

                                                            pipe._increment_stage()

                                                            if pipe.completed:

                                                                transition(obj=pipe,
                                                                           obj_type='Pipeline',
                                                                           new_state=states.DONE,
                                                                           channel=mq_channel,
                                                                           queue='%s-deq-to-sync' % self._sid,
                                                                           profiler=local_prof,
                                                                           logger=self._logger)

                                                        # Found the task and processed it -- no more iterations needed

                                                        break

                                                # Found the stage and processed it -- no more iterations neeeded
                                                break

                                        # Found the pipeline and processed it -- no more iterations neeeded
                                        break

                        mq_channel.basic_ack(delivery_tag=method_frame.delivery_tag)

                    # Appease pika cos it thinks the connection is dead
                    now = time.time()
                    if now - last >= self._rmq_ping_interval:
                        mq_connection.process_data_events()
                        last = now

                except Exception, ex:
                    self._logger.error('Unable to receive message from completed queue: %s' % ex)
                    raise

            self._logger.info('Terminated dequeue thread')
            mq_connection.close()

            local_prof.prof('terminating dequeue-thread', uid=self._uid)

        except KeyboardInterrupt:

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to exit gracefully...')

            mq_connection.close()

            raise KeyboardInterrupt

        except Exception, ex:
            self._logger.exception('Error in dequeue-thread: %s' % ex)

            try:
                mq_connection.close()
            except:
                self._logger.warning('mq_connection not created')

            raise EnTKError(text=ex)

    def _wfp(self):
        """
        **Purpose**: This is the function executed in the wfp process. The function is used to simply create
        and spawn two threads: enqueue, dequeue. The enqueue thread pushes ready tasks to the queues in the pending_q slow
        list whereas the dequeue thread pulls completed tasks from the queues in the completed_q. This function is also
        responsible for the termination of these threads and hence blocking.
        """

        try:

            local_prof = ru.Profiler(name='radical.entk.%s' % self._uid + '-proc', path=self._path)

            local_prof.prof('wfp process started', uid=self._uid)

            self._logger.info('WFprocessor started')

            # Process should run till terminate condtion is encountered
            while (not self._wfp_terminate.is_set()):

                try:

                    # Start dequeue thread
                    if (not self._dequeue_thread) or (not self._dequeue_thread.is_alive()):

                        local_prof.prof('creating dequeue-thread', uid=self._uid)
                        self._dequeue_thread = threading.Thread(
                            target=self._dequeue, args=(local_prof,), name='dequeue-thread')

                        self._logger.info('Starting dequeue-thread')
                        local_prof.prof('starting dequeue-thread', uid=self._uid)
                        self._dequeue_thread.start()

                    # Start enqueue thread
                    if (not self._enqueue_thread) or (not self._enqueue_thread.is_alive()):

                        local_prof.prof('creating enqueue-thread', uid=self._uid)
                        self._enqueue_thread = threading.Thread(
                            target=self._enqueue, args=(local_prof,), name='enqueue-thread')

                        self._logger.info('Starting enqueue-thread')
                        local_prof.prof('starting enqueue-thread', uid=self._uid)
                        self._enqueue_thread.start()

                except Exception, ex:
                    self._logger.error('WFProcessor interrupted')
                    raise

            local_prof.prof('start termination', uid=self._uid)

            self._logger.info('Terminating enqueue-thread')
            self._enqueue_thread_terminate.set()
            self._enqueue_thread.join()
            self._logger.info('Terminating dequeue-thread')
            self._dequeue_thread_terminate.set()
            self._dequeue_thread.join()

            local_prof.prof('termination done', uid=self._uid)

            local_prof.prof('terminating wfp process', uid=self._uid)

            local_prof.close()

        except KeyboardInterrupt:

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to cancel wfprocessor process gracefully...')

            if self._enqueue_thread:

                if not self._enqueue_thread_terminate.is_set():
                    self._logger.info('Terminating enqueue-thread')
                    self._enqueue_thread_terminate.set()
                    self._enqueue_thread.join()

            if self._dequeue_thread:

                if not self._dequeue_thread_terminate.is_set():
                    self._logger.info('Terminating dequeue-thread')
                    self._dequeue_thread_terminate.set()
                    self._dequeue_thread.join()

            self._logger.info('WFprocessor process terminated')

            raise KeyboardInterrupt

        except Exception, ex:
            self._logger.error('Error in wfp process: %s. \n Closing enqueue, dequeue threads' % ex)

            if self._enqueue_thread:

                if not self._enqueue_thread_terminate.is_set():
                    self._logger.info('Terminating enqueue-thread')
                    self._enqueue_thread_terminate.set()
                    self._enqueue_thread.join()

            if self._dequeue_thread:

                if not self._dequeue_thread_terminate.is_set():
                    self._logger.info('Terminating dequeue-thread')
                    self._dequeue_thread_terminate.set()
                    self._dequeue_thread.join()

            self._logger.info('WFprocessor process terminated')

            print traceback.format_exc()
            raise EnTKError(text=ex)

    # ------------------------------------------------------------------------------------------------------------------
    # Public Methods
    # ------------------------------------------------------------------------------------------------------------------

    def start_processor(self):
        """
        **Purpose**: Method to start the wfp process. The wfp function
        is not to be accessed directly. The function is started in a separate
        process using this method.
        """

        if not self._wfp_process:

            try:

                self._prof.prof('creating wfp process', uid=self._uid)
                self._wfp_process = Process(target=self._wfp, name='wfprocessor')

                self._enqueue_thread = None
                self._dequeue_thread = None
                self._enqueue_thread_terminate = threading.Event()
                self._dequeue_thread_terminate = threading.Event()

                self._wfp_terminate = Event()
                self._logger.info('Starting WFprocessor process')
                self._prof.prof('starting wfp process', uid=self._uid)
                self._wfp_process.start()

                return True

            except Exception, ex:

                self._logger.error('WFprocessor not started')
                self.end_processor()
                raise

        else:
            self._logger.warn('Wfp process already running, attempted to restart!')

    def terminate_processor(self):
        """
        **Purpose**: Method to terminate the wfp process. This method is
        blocking as it waits for the wfp process to terminate (aka join).
        """

        try:

            if self.check_processor():
                self._logger.debug('Attempting to end WFprocessor... event: %s' % self._wfp_terminate.is_set())
                self._wfp_terminate.set()
                self._wfp_process.join()
                self._wfp_process = None
                self._logger.debug('WFprocessor process terminated')
            else:
                self._logger.debug('WFprocessor process already terminated')

            self._prof.prof('wfp process terminated', uid=self._uid)

            self._prof.close()

        except Exception, ex:
            self._logger.error('Could not terminate wfprocessor process')
            raise

    def workflow_incomplete(self):
        """
        **Purpose**: Method to check if the workflow execution is incomplete.
        """

        try:
            for pipe in self._workflow:
                with pipe.lock:
                    if pipe.completed:
                        pass
                    else:
                        return True
            return False

        except Exception, ex:
            self._logger.error('Could not check if workflow is incomplete, error:%s' % ex)
            raise

    def check_processor(self):
        """
        **Purpose**: Method to check if the wfp process is alive
        """
        if self._wfp_process:
            return self._wfp_process.is_alive()



__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__ = "Vivek Balasubramanian <vivek.balasubramaniana@rutgers.edu>"
__license__ = "MIT"

import radical.utils as ru
from radical.entk.exceptions import *
import radical.pilot as rp
import os


class Base_ResourceManager(object):

    """
    A resource manager takes the responsibility of placing resource requests on
    different, possibly multiple, DCIs.

    :arguments:
        :resource_desc: dictionary with details of the resource request + access credentials of the user
        :example: resource_desc = {
                                    |  'resource'      : 'xsede.stampede',
                                    |  'walltime'      : 120,
                                    |  'cpus'         : 64,
                                    |  'project'       : 'TG-abcxyz',
                                    |  'queue'         : 'abc',    # optional
                                    |  'access_schema' : 'ssh'  # optional
                                }
    """

    def __init__(self, resource_desc, sid, rts, rts_config):

        if not isinstance(resource_desc, dict):
            raise TypeError(expected_type=dict, actual_type=type(resource_desc))

        self._resource_desc = resource_desc
        self._sid = sid
        self._rts = rts
        self._rts_config = rts_config

        # Resource reservation related parameters
        self._resource = None
        self._walltime = None
        self._cpus = 1
        self._gpus = 0
        self._project = None
        self._access_schema = None
        self._queue = None
        self._validated = False

        # Utility parameters
        self._uid = ru.generate_id('resource_manager.%(item_counter)04d',
                                   ru.ID_CUSTOM,
                                   namespace=self._sid)
        self._path = os.getcwd() + '/' + self._sid
        self._logger = ru.Logger('radical.entk.%s' %
                                 self._uid, path=self._path, targets=['2', '.'])
        self._prof = ru.Profiler(name='radical.entk.%s' % self._uid, path=self._path)

        # Shared data list
        self._shared_data = list()

    @property
    def resource(self):
        """
        :getter: Return user specified resource name
        """
        return self._resource

    @property
    def walltime(self):
        """
        :getter: Return user specified walltime
        """
        return self._walltime

    @property
    def cpus(self):
        """
        :getter: Return user specified number of cpus
        """
        return self._cpus

    @property
    def gpus(self):
        """
        :getter: Return user specified number of gpus
        """
        return self._gpus

    @property
    def project(self):
        """
        :getter: Return user specified project ID
        """
        return self._project

    @property
    def access_schema(self):
        """
        :getter:    Return user specified access schema -- 'ssh' or 'gsissh' or
                    None
        """
        return self._access_schema

    @property
    def queue(self):
        """
        :getter: Return user specified resource queue to be used
        """
        return self._queue

    @property
    def shared_data(self):
        """
        :getter:    list of files to be staged to remote and that are common to
                    multiple tasks
        """
        return self._shared_data

    @shared_data.setter
    def shared_data(self, data_list):

        self._shared_data = data_list

    # ------------------------------------------------------------------------------------------------------------------
    # Public methods
    # ------------------------------------------------------------------------------------------------------------------

    def get_resource_allocation_state(self):
        """
        **Purpose**: Get the state of the resource allocation
        """
        raise NotImplementedError('get_resource_allocation_state() method ' +
                                  'not implemented in ResourceManager for %s' % self._rts)

    def get_completed_states(self):
        """
        **Purpose**: Test if a resource allocation was submitted
        """

        raise NotImplementedError('completed_states() method ' +
                                  'not implemented in ResourceManager for %s' % self._rts)

    # ------------------------------------------------------------------------------------------------------------------
    # Private methods
    # ------------------------------------------------------------------------------------------------------------------

    def _validate_resource_desc(self):
        """
        **Purpose**: Validate the resource description provided to the ResourceManager
        """

        self._prof.prof('validating rdesc', uid=self._uid)
        self._logger.debug('Validating resource description')

        expected_keys = ['resource',
                         'walltime',
                         'cpus']

        for key in expected_keys:
            if key not in self._resource_desc:
                raise MissingError(obj='resource description', missing_attribute=key)

        if not isinstance(self._resource_desc['resource'], str):
            raise TypeError(expected_type=str, actual_type=type(self._resource_desc['resource']))

        if not isinstance(self._resource_desc['walltime'], int):
            raise TypeError(expected_type=int, actual_type=type(self._resource_desc['walltime']))

        if not isinstance(self._resource_desc['cpus'], int):
            raise TypeError(expected_type=int, actual_type=type(self._resource_desc['cpus']))

        if 'gpus' in self._resource_desc:
            if (not isinstance(self._resource_desc['gpus'], int)):
                raise TypeError(expected_type=int, actual_type=type(self._resource_desc['project']))

        if 'project' in self._resource_desc:
            if (not isinstance(self._resource_desc['project'], str)) and (not self._resource_desc['project']):
                raise TypeError(expected_type=str, actual_type=type(self._resource_desc['project']))

        if 'access_schema' in self._resource_desc:
            if not isinstance(self._resource_desc['access_schema'], str):
                raise TypeError(expected_type=str, actual_type=type(self._resource_desc['access_schema']))

        if 'queue' in self._resource_desc:
            if not isinstance(self._resource_desc['queue'], str):
                raise TypeError(expected_type=str, actual_type=type(self._resource_desc['queue']))

        if not isinstance(self._rts_config, dict):
            raise TypeError(expected_type=dict, actual_type=type(self._rts_config))

        self._validated = True

        self._logger.info('Resource description validated')
        self._prof.prof('rdesc validated', uid=self._uid)

        return self._validated

    def _populate(self):
        """
        **Purpose**:    Populate the ResourceManager class with the validated
                        resource description
        """

        if self._validated:

            self._prof.prof('populating rmgr', uid=self._uid)
            self._logger.debug('Populating resource manager object')

            self._resource = self._resource_desc['resource']
            self._walltime = self._resource_desc['walltime']
            self._cpus = self._resource_desc['cpus']
            self._gpus = self._resource_desc.get('gpus', 0)
            self._project = self._resource_desc.get('project', None)
            self._access_schema = self._resource_desc.get('access_schema', None)
            self._queue = self._resource_desc.get('queue', None)

            self._logger.debug('Resource manager population successful')
            self._prof.prof('rmgr populated', uid=self._uid)

        else:
            raise EnTKError('Resource description not validated')

    def _submit_resource_request(self):
        """
        **Purpose**:    Submit resource request as per the description provided by the user
        """

        raise NotImplementedError('_submit_resource_request() method ' +
                                  'not implemented in ResourceManager for %s' % self._rts)

    def _terminate_resource_request(self):
        """
        **Purpose**:    Cancel resource request by terminating any reservation on any acquired
                        resources or resources pending acquisition
        """

        raise NotImplementedError('_terminate_resource_request() method ' +
                                  'not implemented in ResourceManager for %s' % self._rts)



__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__ = "Vivek Balasubramanian <vivek.balasubramanian@rutgers.edu>"
__license__ = "MIT"


# -----------------------------------------------------------------------------
# Possible states
INITIAL = 'DESCRIBED'
SCHEDULING = 'SCHEDULING'
SCHEDULED = 'SCHEDULED'
SUBMITTING = 'SUBMITTING'
SUBMITTED = 'SUBMITTED'
COMPLETED = 'EXECUTED'
DEQUEUEING = 'DEQUEUEING'
DEQUEUED = 'DEQUEUED'
DONE = 'DONE'
FAILED = 'FAILED'
CANCELED = 'CANCELED'

# Final states
FINAL = [DONE, FAILED, CANCELED]

# Assign numeric values to states
state_numbers = {

    INITIAL: 1,
    SCHEDULING: 2,
    SCHEDULED: 3,
    SUBMITTING: 4,
    SUBMITTED: 5,
    COMPLETED: 6,
    DEQUEUEING: 7,
    DEQUEUED: 8,
    DONE: 9,
    FAILED: 9,
    CANCELED: 9
}


# States for Pipeline
_pipeline_state_values = {
    INITIAL: 1,
    SCHEDULING: 2,
    DONE: 9,
    FAILED: 9,
    CANCELED: 9
}

_pipeline_state_inv = {}
for k, v in _pipeline_state_values.iteritems():
    if v in _pipeline_state_inv.keys():
        if not isinstance(_pipeline_state_inv[v], list):
            _pipeline_state_inv[v] = [_pipeline_state_inv[v]]
        _pipeline_state_inv[v].append(k)
    else:
        _pipeline_state_inv[v] = k

# States for Stage
_stage_state_values = {
    INITIAL: 1,
    SCHEDULING: 2,
    SCHEDULED: 3,
    DONE: 9,
    FAILED: 9,
    CANCELED: 9
}

_stage_state_inv = {}
for k, v in _stage_state_values.iteritems():
    if v in _stage_state_inv.keys():
        if not isinstance(_stage_state_inv[v], list):
            _stage_state_inv[v] = [_stage_state_inv[v]]
        _stage_state_inv[v].append(k)
    else:
        _stage_state_inv[v] = k

# States for Task
_task_state_values = {
    INITIAL: 1,
    SCHEDULING: 2,
    SCHEDULED: 3,
    SUBMITTING: 4,
    SUBMITTED: 5,
    COMPLETED: 6,
    DEQUEUEING: 7,
    DEQUEUED: 8,
    DONE: 9,
    FAILED: 9,
    CANCELED: 9
}

_task_state_inv = {}
for k, v in _task_state_values.iteritems():
    if v in _task_state_inv.keys():
        if not isinstance(_task_state_inv[v], list):
            _task_state_inv[v] = [_task_state_inv[v]]
        _task_state_inv[v].append(k)
    else:
        _task_state_inv[v] = k



from radical.entk import Pipeline, Stage, Task, AppManager
import os, sys
from random import shuffle

# ------------------------------------------------------------------------------
# Set default verbosity

if os.environ.get('RADICAL_ENTK_VERBOSE') == None:
    os.environ['RADICAL_ENTK_REPORT'] = 'True'

CUR_NEW_STAGE=0
MAX_NEW_STAGE=4

def generate_pipeline():

    def func_condition():

        global CUR_NEW_STAGE, MAX_NEW_STAGE

        if CUR_NEW_STAGE <= MAX_NEW_STAGE:
            return True

        return False

    def func_on_true():

        global CUR_NEW_STAGE
        CUR_NEW_STAGE += 1
        for t in p.stages[CUR_NEW_STAGE].tasks:
            dur = randint(10,30)
            t.arguments = [str(dur)]

    def func_on_false():
        print 'Done'

    # Create a Pipeline object
    p = Pipeline()

    for s in range(MAX_NEW_STAGE+1):

        # Create a Stage object
        s1 = Stage()

        for i in range(CUR_TASKS):

            t1 = Task()
            t1.executable = ['sleep']
            t1.arguments = [ '30']

            # Add the Task to the Stage
            s1.add_tasks(t1)

        # Add post-exec to the Stage
        s1.post_exec = {
                            'condition': func_condition,
                            'on_true': func_on_true,
                            'on_false': func_on_false
                        }

        # Add Stage to the Pipeline
        p.add_stages(s1)

    return p

if __name__ == '__main__':

    # Create a dictionary describe four mandatory keys:
    # resource, walltime, cores and project
    # resource is 'local.localhost' to execute locally
    res_dict = {

        'resource': 'local.localhost',
        'walltime': 15,
        'cpus': 2,
    }

    # Create Application Manager
    appman = AppManager()
    appman.resource_desc = res_dict

    p = generate_pipeline()

    # Assign the workflow as a set of Pipelines to the Application Manager
    appman.workflow = [p]

    # Run the Application Manager
    appman.run()



__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__ = "Vivek Balasubramanian <vivek.balasubramaniana@rutgers.edu>"
__license__ = "MIT"

import radical.utils as ru
from radical.entk.exceptions import *
import radical.pilot as rp
import os
from ..base.resource_manager import Base_ResourceManager


class ResourceManager(Base_ResourceManager):

    """
    A resource manager takes the responsibility of placing resource requests on
    different, possibly multiple, DCIs. This ResourceManager uses the RADICAL
    Pilot as the underlying runtime system.

    :arguments:
        :resource_desc: dictionary with details of the resource request + access credentials of the user
        :example: resource_desc = {
                                    |  'resource'      : 'xsede.stampede',
                                    |  'walltime'      : 120,
                                    |  'cpus'         : 64,
                                    |  'project'       : 'TG-abcxyz',
                                    |  'queue'         : 'abc',    # optional
                                    |  'access_schema' : 'ssh'  # optional
                                }
    """

    def __init__(self, resource_desc, sid, rts_config):

        super(ResourceManager, self).__init__(resource_desc=resource_desc,
                                              sid=sid,
                                              rts='radical.pilot',
                                              rts_config=rts_config)

        # RP specific parameters
        self._session = None
        self._pmgr = None
        self._pilot = None
        self._download_rp_profile = False

        self._mlab_url = os.environ.get('RADICAL_PILOT_DBURL', None)
        if not self._mlab_url:
            raise EnTKError(msg='RADICAL_PILOT_DBURL not defined. Please assign a valid mlab url')

        if (not "sandbox_cleanup" in self._rts_config.keys()) or (not "db_cleanup" in self._rts_config.keys()):
            raise ValueError(obj=self._uid,
                             attribute='config',
                             expected_value={"sandbox_cleanup": False, "db_cleanup": False},
                             actual_value=self._rts_config)

        self._logger.info('Created resource manager object: %s' % self._uid)
        self._prof.prof('rmgr obj created', uid=self._uid)
    # ------------------------------------------------------------------------------------------------------------------
    # Getter methods
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def session(self):
        """
        :getter: Return reference to the Radical Pilot session instance currently being used
        """
        return self._session

    @property
    def pmgr(self):
        """
        :getter: Return reference to the Radical Pilot manager currently being used
        """
        return self._pmgr

    @property
    def pilot(self):
        """
        :getter: Return reference to the submitted Pilot
        """
        return self._pilot
    # ------------------------------------------------------------------------------------------------------------------
    # Public methods
    # ------------------------------------------------------------------------------------------------------------------

    def get_resource_allocation_state(self):
        """
        **Purpose**: Get the state of the resource allocation

        """

        if self._pilot:
            return self._pilot.state
        else:
            return None

    def get_completed_states(self):
        """
        **Purpose**: Test if a resource allocation was submitted

        """

        return [rp.CANCELED, rp.FAILED, rp.DONE]

    # ------------------------------------------------------------------------------------------------------------------
    # Private methods
    # ------------------------------------------------------------------------------------------------------------------

    def _submit_resource_request(self):
        """
        **Purpose**: Create and submits a RADICAL Pilot Job as per the user
                     provided resource description
        """

        try:

            self._prof.prof('creating rreq', uid=self._uid)

            def _pilot_state_cb(pilot, state):
                self._logger.info('Pilot %s state: %s' % (pilot.uid, state))

                if state == rp.FAILED:
                    self._logger.error('Pilot has failed')

                elif state == rp.DONE:
                    self._logger.error('Pilot has completed')

            self._session = rp.Session(dburl=self._mlab_url, uid=self._sid)
            self._pmgr = rp.PilotManager(session=self._session)
            self._pmgr.register_callback(_pilot_state_cb)

            pd_init = {
                'resource': self._resource,
                'runtime': self._walltime,
                'cores': self._cpus,
                'project': self._project,
            }

            if self._gpus:
                pd_init['gpus'] = self._gpus

            if self._access_schema:
                pd_init['access_schema'] = self._access_schema

            if self._queue:
                pd_init['queue'] = self._queue

            if self._rts_config.get('sandbox_cleanup', None):
                pd_init['cleanup'] = True

            # Create Compute Pilot with validated resource description
            pdesc = rp.ComputePilotDescription(pd_init)

            self._prof.prof('rreq created', uid=self._uid)

            # Launch the pilot
            self._pilot = self._pmgr.submit_pilots(pdesc)

            self._prof.prof('rreq submitted', uid=self._uid)

            shared_staging_directives = list()
            for data in self._shared_data:
                temp = {
                    'source': data,
                    'target': 'pilot:///' + os.path.basename(data)
                }
                shared_staging_directives.append(temp)

            self._pilot.stage_in(shared_staging_directives)

            self._prof.prof('shared data staging initiated', uid=self._uid)
            self._logger.info('Resource request submission successful.. waiting for pilot to go Active')

            # Wait for pilot to go active
            self._pilot.wait([rp.PMGR_ACTIVE, rp.FAILED])

            self._prof.prof('resource active', uid=self._uid)
            self._logger.info('Pilot is now active')

        except KeyboardInterrupt:

            if self._session:
                self._session.close()

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to exit callback thread gracefully...')
            raise KeyboardInterrupt

        except Exception, ex:
            self._logger.error('Resource request submission failed')
            raise

    def _terminate_resource_request(self):
        """
        **Purpose**: Cancel the RADICAL Pilot Job
        """

        try:

            if self._pilot:

                self._prof.prof('canceling resource allocation', uid=self._uid)
                self._pilot.cancel()
                download_rp_profile = os.environ.get('RADICAL_PILOT_PROFILE', False)
                self._session.close(cleanup=self._rts_config.get('db_cleanup', False),
                                    download=download_rp_profile)
                self._prof.prof('resource allocation canceled', uid=self._uid)

        except KeyboardInterrupt:

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to exit callback thread gracefully...')
            raise KeyboardInterrupt

        except Exception, ex:
            self._logger.error('Could not cancel resource request, error: %s' % ex)
            raise

    # ------------------------------------------------------------------------------------------------------------------



#!/usr/bin/env python
__copyright__ = 'Copyright 2013-2018, http://radical.rutgers.edu'
__license__ = 'MIT'


import os
import sys
import glob
import pprint
import radical.utils as ru
import radical.entk as re
import radical.analytics as ra

"""
This example illustrates hoq to obtain durations for arbitrary (non-state)
profile events.
Modified from examples under RADICAL Analytics
"""

# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    loc = './re.session.two.vivek.017759.0012'
    src = os.path.dirname(loc)
    sid = os.path.basename(loc)
    session = ra.Session(src=src, sid = sid, stype='radical.entk')

    # A formatting helper before starting...
    def ppheader(message):
        separator = '\n' + 78 * '-' + '\n'
        print separator + message + separator

    # First we look at the *event* model of our session.  The event model is
    # usually less stringent than the state model: not all events will always be
    # available, events may have certain fields missing, they may be recorded
    # multiple times, their meaning may slightly differ, depending on the taken
    # code path.  But in general, these are the events available, and their
    # relative ordering.
    ppheader("event models")
    pprint.pprint(session.describe('event_model'))
    pprint.pprint(session.describe('statistics'))

    # Let's say that we want to see how long EnTK took to schedule, execute, and
    # process completed tasks.

    # We first filter our session to obtain only the task objects
    tasks = session.filter(etype='task', inplace=False)
    print '#tasks   : %d' % len(tasks.get())

    # We use the 're.states.SCHEDULING' and 're.states.SUBMITTED' probes to find
    # the time taken by EnTK to create and submit all tasks for execution
    ppheader("Time spent to create and submit the tasks")
    duration = tasks.duration(event=[{ru.EVENT: 'state',
                                    ru.STATE: re.states.SCHEDULING},
                                    {ru.EVENT: 'state',
                                    ru.STATE: re.states.SUBMITTED}])
    print 'duration : %.2f' % duration

    # We use the 're.states.SUBMITTED' and 're.states.COMPLETED' probes to find
    # the time taken by EnTK to execute all tasks
    ppheader("Time spent to execute the tasks")
    duration = tasks.duration(event=[{ru.EVENT: 'state',
                                    ru.STATE: re.states.SUBMITTED},
                                    {ru.EVENT: 'state',
                                    ru.STATE: re.states.COMPLETED}])
    print 'duration : %.2f' % duration

    # We use the 're.states.COMPLETED' and 're.states.DONE' probes to find
    # the time taken by EnTK to process all executed tasks
    ppheader("Time spent to process executed tasks")
    duration = tasks.duration(event=[{ru.EVENT: 'state',
                                    ru.STATE: re.states.COMPLETED},
                                    {ru.EVENT: 'state',
                                    ru.STATE: re.states.DONE}])
    print 'duration : %.2f' % duration

    # Finally, we produce a list of the number of concurrent tasks between
    # states 're.states.SUBMITTED' and 're.states.COMPLETED' over the course
    # of the entire execution sampled every 10 seconds
    ppheader("concurrent tasks in between SUBMITTED and EXECUTED states")
    concurrency = tasks.concurrency(event=[{ru.EVENT: 'state',
                                            ru.STATE: re.states.SUBMITTED},
                                            {ru.EVENT: 'state',
                                            ru.STATE: re.states.COMPLETED}],
                                    sampling=10)
    pprint.pprint(concurrency)


# ------------------------------------------------------------------------------
